﻿USE tempdb;
GO

-- jak możemy zapisać CTE?

-- pokaż sprzedaż pracowników
;WITH Dane AS (
    SELECT
		Pracownik		= p.nazwa,
		Produkt			= pr.nazwa,
		RokMiesiac		= YEAR(s.data_sprzedazy) * 100 + MONTH(s.data_sprzedazy),
		s.ilosc,
		s.wartosc
FROM Sprzedaz s

	INNER JOIN Pracownik p
		ON p.id = s.pracownik_id
	
	INNER JOIN Produkt pr
		ON pr.id = s.produkt_id
)
SELECT
	*
FROM Dane

SELECT
	*
FROM (
    SELECT
		Pracownik		= p.nazwa,
		Produkt			= pr.nazwa,
		RokMiesiac		= YEAR(s.data_sprzedazy) * 100 + MONTH(s.data_sprzedazy),
		s.ilosc,
		s.wartosc
FROM Sprzedaz s

	INNER JOIN Pracownik p
		ON p.id = s.pracownik_id
	
	INNER JOIN Produkt pr
		ON pr.id = s.produkt_id
)


-- wersja z nazwami kolumn przed definicją
;WITH Dane (Pracownik, Produkt, RokMiesiąc, Ilość, Wartość) AS (
    SELECT
		Pracownik		= p.nazwa,
		Produkt			= pr.nazwa,
		RokMiesiac		= YEAR(s.data_sprzedazy) * 100 + MONTH(s.data_sprzedazy),
		s.ilosc,
		s.wartosc
FROM Sprzedaz s

	INNER JOIN Pracownik p
		ON p.id = s.pracownik_id
	
	INNER JOIN Produkt pr
		ON pr.id = s.produkt_id
)
SELECT
	*
FROM Dane

-- --------------------
-- A rekurencja?
-- --------------------

-- najpierw pokaż kto jest pod kim
SELECT
	*
FROM Pracownik p

	INNER JOIN Pracownik AS p2
		ON p.id = p2.przelozony_id

-- ok, to spróbujmy pokazać podległość wszystkich pracowników pod prezesem
-- albo chociaż pod jednym z dyrektorów

-- wielokrotne CTE
;WITH Poziom1 AS (
	SELECT
		poziom			= 1,
		sciezka			= ' / ' + p.nazwa,
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p
	WHERE
		p.przelozony_id IS NULL			-- <-- Prezes
		--p.id = 2						-- <-- Dyrektor D1
),
Poziom2 AS (
	SELECT
		poziom			= p1.poziom + 1,		-- lub na sztywno 2
		sciezka			= p1.sciezka + ' / ' + p.nazwa,
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p

		INNER JOIN Poziom1 p1
			ON p1.pracownik_id = p.przelozony_id
),
Poziom3 AS (
	SELECT
		poziom			= p2.poziom + 1,		-- lub na sztywno 3
		sciezka			= p2.sciezka + ' / ' + p.nazwa,
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p

		INNER JOIN Poziom2 p2
			ON p2.pracownik_id = p.przelozony_id
),
Poziom4 AS (
	SELECT
		poziom			= p3.poziom + 1,		-- lub na sztywno 4
		sciezka			= p3.sciezka + ' / ' + p.nazwa,
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p

		INNER JOIN Poziom3 p3
			ON p3.pracownik_id = p.przelozony_id
)

SELECT * FROM Poziom1
UNION ALL
SELECT * FROM Poziom2
UNION ALL
SELECT * FROM Poziom3
UNION ALL
SELECT * FROM Poziom4

-- to mocno upierdliwe jest, bo musimy znać liczbę poziomów (albo dać nadmiarowo dla świętego spokoju)

-- Ale - widzimy, dokąd to zaczyna zmierzać, to może da się zapisać jakoś prościej?
;WITH Poziomy AS (
	-- najpierw dajemy pierwszy warunek; tzw. "kotwica"
	-- ale ponieważ ścieżka się będzie wydłużać nie wiemy jak długo
	-- to od razu musimy dać dłuższy typ danych, to niech będzie VARCHAR(MAX) /a co!/
	-- (dajemy od razu w kotwicy, bo oba fragmenty CTE muszą mieć te same typy danych
	-- a dzięki temu typ będzie przenoszony na kolejne poziomy wywołania)
	SELECT
		poziom			= 1,
		sciezka			= CAST(' / ' + p.nazwa AS VARCHAR(MAX)),
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p
	WHERE
		przelozony_id IS NULL

	-- na koniec robiliśmy UNION ALL wszystkich zbiorów, to dajemy i tu
	UNION ALL

	-- każdy kolejny poziom odwoływał się o jeden wyżej
	-- dla rekurencyjnego CTE jednym poziomem wyżej jest zawsze 
	-- nazwa tego rekurencyjnego CTE, czyli dla nas - "Poziomy"
	-- przez to nie musimy za każdym razem zmieniać na Poziom2, Poziom3, ...
	-- rekurencyjne CTE wie to za nas, a dodatkowo nie musi znać ilości poziomów
	-- bo wykonuje się do momentu, aż zwracane są dane *)
	SELECT
		poziom			= po.poziom + 1,
		sciezka			= po.sciezka + '/' + p.nazwa, --CAST(po.sciezka + '/' + p.nazwa AS VARCHAR(MAX)), -- możemy, ale nie musimy
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p

		INNER JOIN Poziomy po
			ON po.pracownik_id = p.przelozony_id
)

SELECT *
FROM Poziomy

-- *) no, prawie - są wyjątki


-- możemy też zrobić ASCII ART!
;WITH Poziomy AS (
	SELECT
		poziom			= 1,
		sciezka			= CAST(' / ' + p.nazwa AS VARCHAR(MAX)),
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p
	WHERE
		przelozony_id IS NULL

	UNION ALL

	SELECT
		poziom			= po.poziom + 1,
		sciezka			= CAST(po.sciezka + '/' + p.nazwa AS VARCHAR(MAX)),
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p

		INNER JOIN Poziomy po
			ON po.pracownik_id = p.przelozony_id
)

SELECT
	--REPLICATE('  |  ', poziom - 1) + nazwa
	REPLICATE('     ', poziom - 1) + '''-- ' + nazwa
FROM Poziomy
ORDER BY
	sciezka

-- odpytajmy nie tylko od prezesa w dół, ale np. od dyrektora
DECLARE @pracownik_id	INT = 2 -- sprawdzamy wartości 20 oraz 2

;WITH Poziomy AS (
	SELECT
		poziom			= 1,
		sciezka			= CAST(' / ' + p.nazwa AS VARCHAR(MAX)),
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p
	WHERE
		p.id = @pracownik_id

	UNION ALL

	SELECT
		poziom			= po.poziom + 1,
		sciezka			= CAST(po.sciezka + '/' + p.nazwa AS VARCHAR(MAX)),
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p

		INNER JOIN Poziomy po
			ON po.pracownik_id = p.przelozony_id
)

SELECT
	*
FROM Poziomy
ORDER BY
	sciezka

-- co lepsze - wielokrotnie złożone CTE, czy rekurencyjne CTE?
-- Sprawdź plany zapytań obu podejść - zaznacz polecenie powyżej i poniżej

;WITH Poziom1 AS (
	SELECT
		poziom			= 1,
		sciezka			= ' / ' + p.nazwa,
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p
	WHERE
		p.id = @pracownik_id						-- <-- Dyrektor D1
),
Poziom2 AS (
	SELECT
		poziom			= p1.poziom + 1,		-- lub na sztywno 2
		sciezka			= p1.sciezka + ' / ' + p.nazwa,
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p

		INNER JOIN Poziom1 p1
			ON p1.pracownik_id = p.przelozony_id
),
Poziom3 AS (
	SELECT
		poziom			= p2.poziom + 1,		-- lub na sztywno 3
		sciezka			= p2.sciezka + ' / ' + p.nazwa,
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p

		INNER JOIN Poziom2 p2
			ON p2.pracownik_id = p.przelozony_id
),
Poziom4 AS (
	SELECT
		poziom			= p3.poziom + 1,		-- lub na sztywno 4
		sciezka			= p3.sciezka + ' / ' + p.nazwa,
		pracownik_id	= p.id,
		nazwa			= p.nazwa,
		przelozony_id	= p.przelozony_id
	FROM Pracownik p

		INNER JOIN Poziom3 p3
			ON p3.pracownik_id = p.przelozony_id
)

SELECT * FROM Poziom1
UNION ALL
SELECT * FROM Poziom2
UNION ALL
SELECT * FROM Poziom3
UNION ALL
SELECT * FROM Poziom4

-- zwróć uwagę na plan zapytania wielokrotnego CTE - widać fakt, że 
-- odwołanie do CTE jest za każdym razem budowane od nowa!