﻿USE tempdb;
GO

-- Jak ograniczać wynik rekurencji?
-- Sumujemy kolejne liczby od 1 do X

-- domyślnie
;WITH Suma AS (

	SELECT
		liczba	= 1,
		razem	= 1

	UNION ALL

	SELECT
		liczba	= liczba + 1,
		razem	= razem + liczba + 1
	FROM Suma
)

SELECT *
FROM Suma




-- nie za dobrze, dostaliśmy komunikat
-- Msg 530, Level 16, State 1, Line 8
-- The statement terminated. The maximum recursion 100 has been exhausted before statement completion.

-- ograniczmy np. do 25
;WITH Suma AS (

	SELECT
		liczba	= 1,
		razem	= 1

	UNION ALL

	SELECT
		liczba	= liczba + 1,
		razem	= razem + liczba + 1
	FROM Suma
	WHERE
		1 = 1
		-- który warunek jest prawidłowy?
		--AND liczba < 25
		AND liczba <= 25
		--AND liczba < 24
		--AND liczba <= 24
)

SELECT *
FROM Suma
WHERE
	1 = 1
--	-- który warunek jest prawidłowy?
--	AND liczba < 25
--	AND liczba <= 25
--	AND liczba < 24
--	AND liczba <= 24

--OPTION (MAXRECURSION 25) -- inna opcja ograniczenia
OPTION (MAXRECURSION 24) -- powinno być 25, czy 24?












-- odpowiedź:
--AND liczba < 25			-- znak mniejszości, bo mamy też warunek początkowy (dla WHERE w CTE)
							-- żaden (dla WHERE poza CTE) - bo najpierw będzie błąd w rekurencji "The maximum recursion 100 has been exhausted"
--OPTION (MAXRECURSION 24) -- 24, bo mamy warunek początkowy

-- a jak zrobić z liczbą ponad 100?
;WITH Suma AS (

	SELECT
		liczba	= 1,
		razem	= 1

	UNION ALL

	SELECT
		liczba	= liczba + 1,
		razem	= razem + liczba + 1
	FROM Suma
	WHERE
		1 = 1
		AND liczba < 150000
)

SELECT *
FROM Suma
--OPTION (MAXRECURSION 200)
--OPTION (MAXRECURSION 10000)
--OPTION (MAXRECURSION 30000)
--OPTION (MAXRECURSION 40000)
--OPTION (MAXRECURSION 50000)
OPTION (MAXRECURSION 151000)





-- jak to obejść?
;WITH Suma AS (

	SELECT
		liczba	= 1,
		razem	= 1

	UNION ALL

	SELECT
		liczba	= liczba + 1,
		razem	= razem + liczba + 1
	FROM Suma
	WHERE
		1 = 1
		AND liczba < 50000
)

SELECT MAX(liczba), MAX(razem)
FROM Suma
OPTION (MAXRECURSION 0)


-- a co jeśli zrobimy nieograniczoną rekurencję?
;WITH Suma AS (

	SELECT
		liczba	= 1,
		razem	= 1

	UNION ALL

	SELECT
		liczba	= liczba + 1,
		razem	= razem + liczba + 1
	FROM Suma
)

SELECT *
FROM Suma
OPTION (MAXRECURSION 0)










--
-- 65535 rekordów
-- Msg 8115, Level 16, State 2, Line 69
-- Arithmetic overflow error converting expression to data type int.


-- to BIGINT
;WITH Suma AS (

	SELECT
		liczba	= CAST(1 AS BIGINT),
		razem	= CAST(1 AS BIGINT)

	UNION ALL

	SELECT
		liczba	= liczba + 1,
		razem	= razem + liczba + 1
	FROM Suma
)

SELECT *
FROM Suma
OPTION (MAXRECURSION 0)


-- idzie i idzie i idzie ...

